#!/usr/bin/ruby

require 'RMagick'
include Math
include Magick

class Array
    def any?(&b)
	result = false
	each{|*i| result ||= b.call(*i)}
	result
    end
    def each_pair2
	return if size == 0
	lx,ly = self[-1]
	each_with_index{|(x,y),i|
	    yield(x,y,lx,ly)
	    lx,ly = x,y
	}
    end
    def crop(maxx, maxy)
        cut1 = []
	self.each_pair2 {|x,y,lpx,lpy|
	    cut1.push [0, lpy + (y - lpy) * lpx / (lpx - x)] if (x < 0) != (lpx < 0)
	    cut1.push [x,y] if x >= 0
	}
	cut2 = []
	cut1.each_pair2 {|x,y,lpx,lpy|
	    cut2.push [lpx + (x - lpx) * lpy / (lpy - y), 0] if (y < 0) != (lpy < 0)
	    cut2.push [x,y] if y >= 0
	}
	cut3 = []
	cut2.each_pair2 {|x,y,lpx,lpy|
	    cut3.push [maxx, lpy + (y - lpy) * (lpx - maxx) / (lpx - x)] if (x > maxx) != (lpx > maxx)
	    cut3.push [x,y] if x <= maxx
	}
	cut4 = []
	cut3.each_pair2 {|x,y,lpx,lpy|
	    cut4.push [lpx + (x - lpx) * (lpy - maxy) / (lpy - y), maxy] if (y > maxy) != (lpy > maxy)
	    cut4.push [x,y] if y <= maxy
	}
	cut4
    end
end

PI_180 = PI / 180.0

module Projection
    class Cylindrical
	attr_reader :cmd,:cache_key,:rivers_cache_key
	attr_reader :shape

	def initialize(xsize,ysize,north,south,west,east)
    	    @cmd = "#{cname} #{xsize} #{ysize} #{north} #{south} #{west} #{east}"
    	    @cache_key= "#{cname}_#{xsize}x#{ysize}_#{north}_#{south}_#{west}_#{east}"
	    @rivers_cache_key = "#{north}_#{south}_#{west}_#{east}"

	    east  -= 360 if east >=  180
	    west  += 360 if west <  -180
	    east  += 360 if east <= west
	    @xsize = xsize
	    @ysize = ysize
	    @north = north
	    @south = south
	    @west  = west
	    @east  = east
	    @xmul  = @xsize / (@east  - @west)
	    @xadd  = -@west
	    @shape = "rectangle"
	end
	def point_on_map?(x,y)
	    x >= @west && x <= @east && y >= @south && y <= @north
	end
	def bb_on_map?(l,r,u,d)
	    r >= @west && l <= @east && u >= @south && d <= @north
	end
	def map_intersect(b)
	    Box.new([@west, b.l].max, [@east, b.r].min, [@north, b.u].min, [@south, b.d].max)
	end
	def projected_box_area(box)
	    x1,y1 = project(box.l, box.u)
	    x2,y2 = project(box.r, box.d)
	    ((x1-x2) * (y1-y2)).abs
	end
    end
    class Azimuthal
	attr_reader :cmd,:cache_key,:rivers_cache_key
	attr_reader :shape

	def initialize(radius,center_x,center_y)
	    @xsize  = 2 * radius
	    @ysize  = 2 * radius

	    if center_y == 0
		@north  =  90.0
		@south  = -90.0
	        @west   = center_x - 90
	        @east   = center_x + 90
	    elsif center_y == 90
		@north  =  90.0
		@south  =   0.0
	        @west   = center_x - 180
	        @east   = center_x + 180
	    elsif center_y == -90
		@north  =   0.0
		@south  = -90.0
	        @west   = center_x - 180
	        @east   = center_x + 180
	    else
		raise "Only equator-centered and pole-centered azimuthal maps are supported"
	    end

	    @west  += 360 if @west <  -180
	    @east  += 360 if @east <= @west
	
	    @long0  = center_x * PI_180
	    @lat0   = center_y * PI_180
	    @radius = radius
	    @shape  = "round"

    	    @cache_key= "azhemi_#{@xsize}x#{@ysize}_#{@north}_#{@south}_#{@west}_#{@east}"
	    @rivers_cache_key = "#{@north}_#{@south}_#{@west}_#{@east}"
	end
	def point_on_map?(x,y)
	    x >= @west && x <= @east && y >= @north && y <= @south
	end
	def bb_on_map?(l,r,u,d)
	    r >= @west && l <= @east && u >= @south && d <= @north
	end
	def map_intersect(b)
	    Box.new([@west, b.l].max, [@east, b.r].min, [@north, b.u].min, [@south, b.d].max)
	end
	def project(long,lat)
	    long = @west  if long < @west
	    long = @east  if long > @east
	    lat  = @north if lat  > @north
	    lat  = @south if lat  < @south

	    long *= PI_180
	    lat  *= PI_180

	    long -= @long0
	    
	    g = sin(@lat0) * sin(lat) + cos(@lat0) * cos(lat) * cos(long)
	    raise "hmmm" if (g == -1)
	    ksp = @radius * sqrt (1.0 / (1.0 + g))
	    x = ksp *  cos (lat) * sin(long) + @xsize/2
	    y = - ksp * (cos(@lat0) * sin(lat) - sin(@lat0) * cos(lat) * cos(long)) + @ysize/2
	    
	    [x,y]
	end
    end
    class Equidistant < Cylindrical
	def initialize(xsize,ysize,north,south,west,east)
	    super(xsize,ysize,north,south,west,east)
	    @ymul = @ysize / (@south - @north)
	    @yadd = -@north
	end
	def project(long,lat)
	    [(long + @xadd) * @xmul,
	     (lat  + @yadd) * @ymul
	    ]
	end
	def cname
	    "d"
	end
    end
    class Equalarea < Cylindrical
	def initialize(xsize,ysize,north,south,west,east)
	    super(xsize,ysize,north,south,west,east)
	    @ymul = @ysize / (sin (@south * PI_180) - sin(@north * PI_180))
	    @yadd = -sin(@north * PI_180)
	end
	def project(long,lat)
	    [(long + @xadd) * @xmul,
	     (sin (lat * PI_180) + @yadd) * @ymul
	    ]
	end
	def cname
	    "a"
	end
    end
    class Mercator < Cylindrical
	def forw(lat)
	    lat = (89 + lat)/2 if lat > 89
	    lat = (-89 + lat)/2 if lat < -89
	    log(tan(lat * PI_180) + 1/cos(lat * PI_180))
	end
	def project(long,lat)
	    [(long + @xadd) * @xmul,
	     (forw (lat)    - forw(@north))/
	     (forw (@south) - forw(@north)) * @ysize
	    ]
	end
	def cname
	    "m"
	end
    end
end

class GIS
    def arc
	load_arc unless @arc
	@arc
    end
    def arcbb
	load_arc unless @arcbb
	@arcbb
    end
    def arcr
	load_arc unless @arcr
	@arcr
    end
    def colors
	load_colors unless @colors
	@colors
    end
    def ccolors
	load_ccolors unless @ccolors
	@ccolors
    end
    def barycenters
	load_barycenters unless @barycenters
	@barycenters
    end
    def coutlines
	load_barycenters unless @coutlines
	@coutlines
    end
    def cnames(locale)
	load_cnames(locale) unless @cnames
	@cnames
    end
    def oversea
	load_oversea unless @oversea
	@oversea
    end
    def lab
	load_lab unless @lab
	@lab
    end    
    def load_arc
        @arc=[]
        @arcbb=[]
	@arcr=[]
	open("pp/borders"){|fn|fn.each{|line|
	    line =~ /(\d+) (\S+) (\S+) (\S+) (\S+) (.*)/
    	    @arc.push $6.split(" ")
	    @arcr.push $1.to_i
	    @arcbb.push Box.new($2.to_f, $3.to_f, $5.to_f, $4.to_f)
	}}
    end
    def load_lab
	@lab = []
        open("data/lab"){|fn|fn.each{|line|
	    line =~ /(\d+) (.*)/
    	    @lab[$1.to_i]=$2
	}}    
    end
    def load_colors
	@colors = []
        open("pp/colors"){|fn|fn.each{|line|
    	    line =~ /(\d+) (\S+)/
    	    @colors[$1.to_i]=$2
	}}
    end
    def load_ccolors
	@ccolors = {}
        open("pp/ccolors"){|fn|fn.each{|line|
    	    line =~ /(.+?)\t(\S+)/
    	    @ccolors[$1]=$2.to_i
	}}
    end
    def load_oversea
	@oversea={}
        open("data/oversea"){|fn|fn.each{|line|
	    line =~ /(.*?)\t+(.*)/
    	    @oversea[$1]=$2
	}}
    end
    def load_cnames(locale)
	@cnames = {}
        if locale
	    open("data/clocale.#{locale}"){|fn|fn.each{|line|
		line =~ /(.*?)\t+(.*)/
		@cnames[$1]=$2
	    }}
	end
    end
    def load_barycenters
        @barycenters={}
	@coutlines={}
	country={}
	nr = 0
        open("pp/pal"){|fn|fn.each{|line|
	    line.chomp!

    	    nr += 1
	    next if nr == 1040

	    country_name = lab[nr]
	    next if country_name == "?"

    	    p = line.split (" ").map{|z|z.to_i}
	    box = nil
	    p.each {|num|
	        next if num == 0
		b = arcbb[num.abs - 1]
	        if box
		    box += b
	        else
		    box = b
		end
	    }
    	    normal  = $proj.bb_on_map?(box.l, box.r, box.u, box.d)
	    shifted = $proj.bb_on_map?(box.l + 360, box.r + 360, box.u, box.d)
	    next unless normal or shifted

	    outline_n=[]
	    outline_s=[]

	    nptx,npty = nil,nil
            nholex,nholey = nil,nil
	    sptx,spty = nil,nil
            sholex,sholey = nil,nil

	    catch (:dontrender) { p.each{|num|
		if num == 0
		    throw :dontrender unless (nptx or sptx)
		    if nholex or sholex
			outline_n.push [nholex, nholey] if normal
			outline_s.push [sholex, sholey] if shifted
		    else
		        nholex, nholey = nptx, npty
			sholex, sholey = sptx, spty
		    end
		else
		    if num > 0
			c = arc[num-1]
		    else
	    		c = arc[-num-1].reverse
		    end
		    c.each {|ptd|
			x,y = ptd.split(",")
			if normal
			    nptx,npty = $proj.project(x.to_f, y.to_f)
	    		    outline_n.push [nptx,npty]
			end
		        if shifted
			    sptx,spty = $proj.project(x.to_f + 360, y.to_f)
	    		    outline_s.push [sptx,spty]
			end
		    }
		end
	    }
	    outline_n.push [nholex,nholey] if normal  and nholex
	    outline_s.push [sholex,sholey] if shifted and sholex
	    if normal
		area, acx, acy = 0.0, 0.0, 0.0
		outline_nc = outline_n.crop($config.map_x, $config.map_y)
		outline_nc.each_pair2 {|x,y,lpx,lpy|
		    sf    = lpx * y - x * lpy
	    	    area += sf
		    acx  += (x + lpx) * sf
	    	    acy  += (y + lpy) * sf
		}
		next if area == 0.0
		x0, y0, a0, ol = country[country_name] || [0.0, 0.0, 0.0, nil]
		ol  += outline_nc + [outline_nc[0]] + [ol[-1]] if ol
		ol ||= outline_nc + [outline_nc[0]]
		country[country_name] = [x0 + acx / 6, y0 + acy / 6, a0 + area / 2, ol]
	    end
	    if shifted
		area, acx, acy = 0.0, 0.0, 0.0
		outline_sc = outline_s.crop($config.map_x, $config.map_y)
		outline_sc.each_pair2 {|x,y,lpx,lpy|
		    sf    = lpx * y - x * lpy
	    	    area += sf
		    acx  += (x + lpx) * sf
	    	    acy  += (y + lpy) * sf
		}
		next if area == 0.0
		x0, y0, a0, ol = country[country_name] || [0.0, 0.0, 0.0, nil]
		ol  += outline_sc + [outline_sc[0]] + [ol[-1]] if ol
		ol ||= outline_sc + [outline_sc[0]]
		country[country_name] = [x0 + acx / 6, y0 + acy / 6, a0 + area / 2, ol]
	    end
	}}}
	country.each{|nam,(x,y,a,ol)|
	    @barycenters[nam] = [x/a, y/a, a]
	    @coutlines[nam] = ol
	}
    end
end

class Box < Struct.new(:l,:r,:u,:d)
    def disjoint?(b)
	b.r <= l || b.l >= r || b.u >= d || b.d <= u
    end
    def fully_inside?(b)
	b.l >= l && b.r <= r && b.u <= u && b.d >= d
    end
    def overlap(b)
	[r - b.l, b.r - l].max * [d - b.u, b.d - u].max
    end
    def +(b)
	Box.new([l,b.l].min, [r,b.r].max, [u,b.u].max, [d,b.d].min)
    end
    def mx
	(l + r) / 2
    end
    def my
	(u + d) / 2
    end
end

class CityDot < Struct.new(:x,:y,:overlap_range)
    def box
	@box ||= Box.new(x-overlap_range, x+overlap_range, y-overlap_range, y+overlap_range)
	@box
    end
end

class CityLabel < Struct.new(:rx,:ry,:x,:y,:w,:h,:label,:offmax)
    def box
	l = x + rx * (w / 2 + offmax) - w/2
        r = l + w
        u = y + ry * (h / 2 + offmax) - h/2
        d = u + h
	Box.new (l,r,u,d)
    end
end

class Labels
    def initialize(image,font=nil,color='white')
	@image = image
	@color = color
	@text = Draw.new
	@text.font = font if font
	@text.text_antialias = true
	@text.pointsize = 18
	@text.fill = @color
	@labels = []
	@citydots = []
    end
    def improve_placement(temp)
	buckets=[]
	@labels.each_with_index {|l,i|
	    b = l.box
	    conflicts = []
	    buckets.each_with_index{|(bb,labels),j|
		if not b.disjoint?(bb) && labels.any?{|k| not b.disjoint?(@labels[k].box)}
		    conflicts.push j
		end
	    }
	    lss=[i]
	    dsf=0
	    conflicts.each {|j|
	        bb,ls = buckets.delete_at (j-dsf)
		dsf += 1
		b   += bb
	        lss += ls
	    }
	    buckets.push [b, lss]
	}
	vx=[]
	vy=[]
	buckets.each{|b,ls|
	    ls.each {|i| ls.each{|j|
		next if i == j
		l = @labels[i].box
		q = @labels[j].box
		unless l.disjoint? q
		    ivx = (l.mx - q.mx)
		    ivy = (l.my - q.my)
		    ivn = sqrt(ivx * ivx + ivy * ivy)
		    o = l.overlap q
		    vx[i] ||= 0.0
		    vy[i] ||= 0.0
		    vx[i] += o * ivx / ivn
		    vy[i] += o * ivy / ivn
		end
	    }}
	}
	@labels.each_with_index{|el,i|
	    b = el.box
	    rx = el.rx
	    ry = el.ry
	    vx[i] ||= 0.0
	    vy[i] ||= 0.0
	    @citydots.each{|c|
		unless b.disjoint? c.box
		    ivx = b.mx - c.x
		    ivy = b.mx - c.y
		    ivn = sqrt(ivx * ivx + ivy * ivy)
		    vx[i] += ivx / ivn
		    vy[i] += ivy / ivn
		end
	    }
	    n = sqrt (vx[i] * vx[i] + vy[i] * vy[i])
	    if n == 0
		if rx.abs < ry.abs
		    rx = (rx > 0) ? [rx-temp, 0].max : [rx+temp, 0].min
		else
		    ry = (ry > 0) ? [ry-temp, 0].max : [ry+temp, 0].min
		end
	    else
	        rx = rx + temp * vx[i]/n
		ry = ry + temp * vy[i]/n
		rx =  1 if rx >  1
		rx = -1 if rx < -1
		ry =  1 if ry >  1
		ry = -1 if ry < -1
	    end
	    @labels[i].rx = rx
	    @labels[i].ry = ry
	}
    end
    def draw
	10.times {improve_placement 0.5}
	10.times {improve_placement 0.4}
	10.times {improve_placement 0.3}
	10.times {improve_placement 0.2}
	10.times {improve_placement 0.1}
	10.times {improve_placement 0.05}
	@labels.each {|l|
	    b = l.box
	    @text.annotate(@image, 0, 0, b.l, b.d, l.label)
	}
    end
    def add_city(size,name,long,lat)
	if $proj.point_on_map?(long,lat)
	    x,y = $proj.project (long,lat)
	elsif $proj.point_on_map?(long+360,lat)
	    x,y = $proj.project (long+360,lat)
	else
	    return
	end
	if size == "b"
	    size          = 4
	    overlap_range = 6
	    offmax        = 8
	else
	    size          = 2.5
	    overlap_range = 3
	    offmax        = 5
	end

	dot = Draw.new # antialias
	dot.stroke 'none'
        dot.fill @color
	dot.circle(x, y, x, y + size)
        dot.draw @image
	
	metrics = @text.get_type_metrics(@image,name)
	@labels.push CityLabel.new(1, 0, x, y, metrics.width, metrics.ascent, name, offmax)
	@citydots.push CityDot.new(x, y, overlap_range)
    end
end

class Config
    attr_reader :rivers,:clabels,:borders,:locale,:cities,:font
    attr_reader :proj, :outfile, :type_bm
    attr_reader :map_x,:map_y
    def initialize(fh)
	@rivers       = false
	@clabels      = false
	@borders      = false
	@locale	      = nil
	@cities       = nil
	@font	      = nil
	@outfile      = nil
	@type_bm      = true

	name              = nil
	center_x,center_y = nil,nil
	size_x,size_y     = nil,nil
	extend_x,extend_y = nil,nil
	radius            = nil
	projection        = 'equalarea'

	fh.each{|line|
	    line.chomp!
	    case line
	    when /size\s*=\s*(\d+)\s+(\d+)/
		@map_x = $1.to_i
		@map_y = $2.to_i
	    when /center\s*=\s*(-?\d+\.?\d*)\s+(-?\d+\.?\d*)/
		center_x = $1.to_f
		center_y = $2.to_f
	    when /extend\s*=\s*(-?\d+\.?\d*)\s+(-?\d+\.?\d*)/
		extend_x = $1.to_f
		extend_y = $2.to_f
	    when /radius\s*=\s*(\d+)/
		radius = $1.to_i
		@map_x = radius * 2
		@map_y = radius * 2
	    when /background\s*=\s*(.*)/
		case $1
		when 'satellite','bluemarble'
		    @type_bm = true
		when 'political','polygons'
		    @type_bm = false
		else
	    	    STDERR.print "Unknown map type: #{$1}\n"
		    exit 1
		end
	    when /locale\s*=\s*(.*)/
		@locale = $1
	    when "rivers"
		@rivers = true
	    when "borders"
		@borders = true
	    when "clabels"
		@clabels = true
	    when /cities\s*=\s*(.*)/
		@cities = $1
	    when /font\s*=\s*(.*)/
		@font = $1
	    when /name\s*=\s*(.*)/
		name = $1
	    when /projection\s*=\s*(.*)/
		projection = $1
	    else
		STDERR.print "Unknown line: #{line}\n"
		exit 1
	    end
	}
	unless name
	    STDERR.print "No output file specified\n"
	    exit 1
	end
	version  = 410
	version += 1 while File.exists? "#{name}-#{version}.png"
	@outfile = "#{name}-#{version}.png"
	print "#{name}: #{version}\n"

	case projection
	when 'equalarea','a'
	    @proj = Projection::Equalarea.new  (@map_x, @map_y, center_y + extend_y, center_y - extend_y,
							        center_x - extend_x, center_x + extend_x)
	when 'equidistant','d'
	    @proj = Projection::Equidistant.new(@map_x, @map_y, center_y + extend_y, center_y - extend_y,
							        center_x - extend_x, center_x + extend_x)
	when 'mercator','m'
	    @proj = Projection::Mercator.new   (@map_x, @map_y, center_y + extend_y, center_y - extend_y,
							        center_x - extend_x, center_x + extend_x)
	when 'azimuthal','z'
	    @proj = Projection::Azimuthal.new  (radius, center_x, center_y)
	else
	    STDERR.print "Unknown projection #{$1}\n"
	    exit 1
	end
    end
end

$config = Config.new(STDIN)
$proj = $config.proj
$gis  = GIS.new

if $config.type_bm
    blue_marble_cache_file = "blue_marble_cache/" + $proj.cache_key + ".ppm"
    raise "Blue Marble only supports rectangular maps" unless $proj.shape == "rectangle"
    system ("./project " + $proj.cmd + " >#{blue_marble_cache_file}") unless File.exist? blue_marble_cache_file
    image = ImageList.new(blue_marble_cache_file)
else
    polygon_cache_file = "polygon_cache/" + $proj.cache_key + ".png"
    unless File.exist? polygon_cache_file
        require 'libart'

	ccolors = $gis.ccolors
        oversea = $gis.oversea

	if $proj.shape == "rectangle"
	    ximage  = Art::Canvas.new($config.map_x, $config.map_y, 0xE8E8FFFF)
	else
	    ximage  = Art::Canvas.new($config.map_x, $config.map_y, 0x00000000)
	    circle  = Art::Vpath.circle($config.map_x/2, $config.map_y/2, $config.map_x/2)
	    svp     = circle.to_svp
	    ximage.render_svp(svp, 0xE8E8FFFF)
	end

        $gis.coutlines.each{|country,outline|
	    voutline = outline.map{|x,y| [Art::LINETO, x, y]}
    	    voutline[0][0] = Art::MOVETO if voutline[0][0] == Art::LINETO
	    voutline.push [Art::END]
	    voutline = Art::Vpath.new voutline

	    color = ccolors[oversea[country] || country]
	    svp = voutline.to_svp
	    ximage.render_svp(svp, color)
	}
	File::open(polygon_cache_file, "w") {|fn| fn.write ximage.to_png}
    end
    image = ImageList.new(polygon_cache_file)
end

if $config.borders
    arc   = $gis.arc
    arcbb = $gis.arcbb
    arcr  = $gis.arcr

    arc.each_with_index {|outline,i|
   	next if arcr[i] < 4
        box  = arcbb[i]
    	normal  = $proj.bb_on_map? (box.l, box.r, box.u, box.d)
	shifted = $proj.bb_on_map? (box.l + 360, box.r + 360, box.u, box.d)
    	d = Draw.new
	d.stroke_antialias true
	if $config.type_bm
    	    d.stroke_width 1.5
	    d.stroke "gray60"
	else
	    if $1.to_i == 5
	        d.stroke_width 2
	        d.stroke "#FF4040"
	    else
	        d.stroke_width 1
	        d.stroke "black"
	        d.stroke_opacity "50%"
	    end
	end
	d.fill "none"
	if normal
            outline2 = []
	    outline.each {|xy|
		x,y = xy.split(",")
		x,y = $proj.project(x.to_f, y.to_f)
	    	outline2.push x
		outline2.push y
	    }
    	    d.polyline(*outline2)
	    d.draw image
	end
	if shifted
            outline2 = []
	    outline.each {|xy|
		x,y = xy.split(",")
		x,y = $proj.project(x.to_f + 360, y.to_f)
	    	outline2.push x
		outline2.push y
	    }
    	    d.polyline(*outline2)
	    d.draw image
	end
    }
end

if $config.rivers
    rivers_cache_file = "rivers_cache/" + $proj.rivers_cache_key
    unless File.exists? rivers_cache_file
	rivers_cache_fh = open (rivers_cache_file, "w")
        open("pp/rivers") {|fn| fn.each{|line|
	    line    =~ /(\d+) (\S+) (\S+) (\S+) (\S+) (.*)/
    	    rank    = $1.to_i
	    outline = $6
#        break if rank >= 4
	    if $proj.bb_on_map? ($2.to_f, $3.to_f, $5.to_f, $4.to_f)
		rivers_cache_fh.print rank, " ", outline, "\n"
	    end
	    if $proj.bb_on_map? ($2.to_f + 360, $3.to_f + 360, $5.to_f, $4.to_f)
		rivers_cache_fh.print rank, " "
	        rivers_cache_fh.print outline.split(" ").map {|xy|
		    x,y = xy.split(",")
		    "#{x.to_f+360},#{y}"
    		}.join(" ")
		rivers_cache_fh.print "\n"
	    end
	}}
	rivers_cache_fh.close
    end
    open(rivers_cache_file) {|fn| fn.each{|line|
	line =~ /(\d+) (.*)/
        rank = $1.to_i

        d = Draw.new
	d.stroke_antialias true
        if rank == 1
	    d.stroke_width 3
        elsif rank == 2
	    d.stroke_width 2
        else # rank == 3
	    d.stroke_width 1.5
        end
        if $config.type_bm
	    d.stroke "#0040FF"
	    d.opacity "50%"
        else
	    d.stroke "#8080FF"
        end
        d.fill "none"
        outline = $2.split(" ")
        outline2 = []
        outline.map {|xy|
	    x,y = xy.split(",")
	    x,y = $proj.project(x.to_f, y.to_f)
    	    outline2.push x
	    outline2.push y
        }
        d.polyline(*outline2)
        d.draw image
    }}
end

if $config.cities
    if $config.type_bm
	labels=Labels.new(image,$config.font,"white")
    else
	labels=Labels.new(image,$config.font,"black")
    end
    fn=open($config.cities)
    fn.each {|line|
	line =~ /^(.)\s+(.*?)\s+(-?\d+\.?\d*)\s+(-?\d+\.?\d*)/
        labels.add_city($1, $2, $4.to_f, $3.to_f)
    }
    labels.draw
end

if $config.clabels
    lab           = $gis.lab
    bcentrs       = $gis.barycenters
    oversea       = $gis.oversea
    locale_cnames = $gis.cnames($config.locale)

    bcentrs.each{|country_name,(x,y,a)|
	lcname        = locale_cnames[country_name]  || country_name
	mainland_name = oversea[country_name]
	lmname        = locale_cnames[mainland_name] || mainland_name if mainland_name

        text = Draw.new
	text.font = $config.font if $config.font
        text.text_antialias = true
	if $config.type_bm
	    text.fill = "#B350B350FFFF8000"
	else
	    text.fill = '#800000'
	end
        text.font_weight = BoldWeight
	if a > 200000
    	    pointsize = 72
	elsif a > 100000
    	    pointsize = 48
	elsif a > 30000
	    pointsize = 36
	elsif a > 10000
	    pointsize = 24
	elsif a > 3000
	    pointsize = 18
	elsif a > 100
	    pointsize = 12
#	elsif a > 10
#	    pointsize = 8
	else
	    next
	end
	text.pointsize = pointsize
	if lmname
	    lmname = "(#{lmname})"
	    metrics1 = text.get_type_metrics(image, lcname)
	    metrics2 = text.get_type_metrics(image, lmname)
	    text.annotate(image, 0, 0, x - metrics1.width/2, y, lcname)
	    text.annotate(image, 0, 0, x - metrics2.width/2, y - metrics1.descent + metrics2.ascent, lmname)
	else
	    metrics = text.get_type_metrics(image, lcname)
	    d,u = y + metrics.ascent/2, y - metrics.ascent/2
	    l,r = x - metrics.width/2,  x + metrics.width/2
	    if l < 5
		l = 5
		r = l + metrics.width
	    elsif r > ($config.map_x - 5)
		r = ($config.map_x - 5)
		l = r - metrics.width
	    end
	    if u < 5
		u = 5
		d = l + metrics.ascent
	    elsif d > ($config.map_y - 5)
		d = ($config.map_y - 5)
		u = d - metrics.ascent
	    end
	    text.annotate(image, 0, 0, l, d, lcname)
	end
    }
end

image.write($config.outfile)
